# Copyright (c) 2023-2024 Datalayer, Inc.
#
# Datalayer License

"""Run handler."""

import tornado
from jupyter_server.base.handlers import APIHandler
from jupyter_server.extension.handler import ExtensionHandlerMixin

from ...__version__ import __version__


# pylint: disable=W0223
class ContentHandler(ExtensionHandlerMixin, APIHandler):
    """The content handler."""

    @tornado.web.authenticated
    def get(self):
        """Returns the content."""
        pass



# Copyright (c) 2023-2024 Datalayer, Inc.
#
# Datalayer License

"""Index handler."""

import tornado

from ..base import BaseTemplateHandler


# pylint: disable=W0223
class IndexHandler(BaseTemplateHandler):
    """The handler for the index."""

    @tornado.web.authenticated
    def get(self):
        """The index page."""
        self.write(self.render_template("index.html"))



# Copyright (c) 2023-2024 Datalayer, Inc.
#
# Datalayer License

"""Base handler."""

from jupyter_server.base.handlers import JupyterHandler
from jupyter_server.extension.handler import ExtensionHandlerMixin, ExtensionHandlerJinjaMixin


# pylint: disable=W0223
class BaseTemplateHandler(ExtensionHandlerJinjaMixin, ExtensionHandlerMixin, JupyterHandler):
    """The Base handler for the templates."""
    


# Copyright (c) 2023-2024 Datalayer, Inc.
#
# Datalayer License

import sys
import warnings
from pathlib import Path

from traitlets import Bool, Unicode

from jupyter_kernels.__version__ import __version__

from datalayer_core.application import (
    DatalayerApp, NoStart, base_aliases, base_flags
)


HERE = Path(__file__).parent


jupyter_iam_aliases = dict(base_aliases)
jupyter_iam_aliases["cloud"] = "JupyterKernelsApp.cloud"

jupyter_iam_flags = dict(base_flags)
jupyter_iam_flags["dev-build"] = (
    {"JupyterKernelsApp": {"dev_build": True}},
    "Build in development mode.",
)
jupyter_iam_flags["no-minimize"] = (
    {"JupyterKernelsApp": {"minimize": False}},
    "Do not minimize a production build.",
)


class ConfigExportApp(DatalayerApp):
    """An application to export the configuration."""

    version = __version__
    description = """
   An application to export the configuration
    """

    def initialize(self, *args, **kwargs):
        """Initialize the app."""
        super().initialize(*args, **kwargs)

    def start(self):
        """Start the app."""
        if len(self.extra_args) > 1:  # pragma: no cover
            warnings.warn("Too many arguments were provided for workspace export.")
            self.exit(1)
        self.log.info("JupyterKernelsConfigApp %s", self.version)


class JupyterKernelsConfigApp(DatalayerApp):
    """A config app."""

    version = __version__
    description = """
    Manage the configuration
    """

    subcommands = {}
    subcommands["export"] = (
        ConfigExportApp,
        ConfigExportApp.description.splitlines()[0],
    )

    def start(self):
        try:
            super().start()
            self.log.error("One of `export` must be specified.")
            self.exit(1)
        except NoStart:
            pass
        self.exit(0)


class JupyterKernelsShellApp(DatalayerApp):
    """A shell app."""

    version = __version__
    description = """
    Run predefined scripts.
    """

    def start(self):
        super().start()
        args = sys.argv
        self.log.info(args)


class JupyterKernelsApp(DatalayerApp):
    name = "jupyter_iam"
    description = """
    Import or export a JupyterLab workspace or list all the JupyterLab workspaces

    You can use the "config" sub-commands.
    """
    version = __version__

    aliases = jupyter_iam_aliases
    flags = jupyter_iam_flags

    minimize = Bool(True, config=True, help="")

    subcommands = {
        "config": (JupyterKernelsConfigApp, JupyterKernelsConfigApp.description.splitlines()[0]),
        "sh": (JupyterKernelsShellApp, JupyterKernelsShellApp.description.splitlines()[0]),
    }

    def initialize(self, argv=None):
        """Subclass because the ExtensionApp.initialize() method does not take arguments"""
        super().initialize()

    def start(self):
        super(JupyterKernelsApp, self).start()
        self.log.info("Jupyter Kernels - Version %s", self.version)


# -----------------------------------------------------------------------------
# Main entry point
# -----------------------------------------------------------------------------

main = launch_new_instance = JupyterKernelsApp.launch_instance

if __name__ == "__main__":
    main()



# Copyright (c) 2023-2024 Datalayer, Inc.
#
# Datalayer License

from typing import Any, Dict, List

from .__version__ import __version__
from .serverapplication import JupyterKernelsExtensionApp


def _jupyter_server_extension_points() -> List[Dict[str, Any]]:
    return [{
        "module": "jupyter_kernels",
        "app": JupyterKernelsExtensionApp,
    }]



# Copyright (c) 2023-2024 Datalayer, Inc.
#
# Datalayer License

import json

from ..__version__ import __version__


async def test_config(jp_fetch):
    # When
    response = await jp_fetch("jupyter_kernels", "config")
    # Then
    assert response.code == 200
    payload = json.loads(response.body)
    assert payload == {
        "extension": "jupyter_kernels",
        "version": __version__
    }



# Copyright (c) 2023-2024 Datalayer, Inc.
#
# Datalayer License

"""The Jupyter Kernels Server application."""

import os

from traitlets import CInt

from jupyter_server.utils import url_path_join
from jupyter_server.extension.application import ExtensionApp, ExtensionAppJinjaMixin

from jupyter_kernels.handlers.index.handler import IndexHandler
from jupyter_kernels.handlers.config.handler import ConfigHandler
from jupyter_kernels.handlers.jump.handler import JumpHandler
from jupyter_kernels.handlers.content.handler import ContentHandler
# from .handlers.jump.keys import setup_keys


from jupyter_kernels.__version__ import __version__


DEFAULT_STATIC_FILES_PATH = os.path.join(os.path.dirname(__file__), "./static")

DEFAULT_TEMPLATE_FILES_PATH = os.path.join(
    os.path.dirname(__file__), "./templates")


class JupyterKernelsExtensionApp(ExtensionAppJinjaMixin, ExtensionApp):
    """The Jupyter Kernels Server extension."""

    name = "jupyter_kernels"

    extension_url = "/jupyter_kernels"

    load_other_extensions = True

    static_paths = [DEFAULT_STATIC_FILES_PATH]
    template_paths = [DEFAULT_TEMPLATE_FILES_PATH]

    max_notebook_kernels = CInt(
        5,
        config=True,
        help="Maximal number of Notebook Remote Kernels that can be spawned by an user."
    )

    max_cell_kernels = CInt(
        3,
        config=True,
        help="Maximal number of Cell Remote Kernels that can be spawned by an user. They are not taking into account in the max_notebook_kernels limit."
    )

    def initialize_settings(self):
        settings = dict(
            max_notebook_kernels=self.max_notebook_kernels,
            max_cell_kernels=self.max_cell_kernels
        )
        self.settings.update(**settings)

    def initialize_templates(self):
        self.serverapp.jinja_template_vars.update(
            {
                "jupyter_kernels_version": __version__,
            })

    def initialize_handlers(self):
        pod_name_regex = r"(?P<pod_name>[\w\.\-%]+)"
        handlers = [
            ("jupyter_kernels", IndexHandler),
            (url_path_join("jupyter_kernels", "config"), ConfigHandler),
            (url_path_join("jupyter_kernels", "content"), ContentHandler),
            (r"/jupyter_kernels/jump/%s" % pod_name_regex, JumpHandler),
        ]
        self.handlers.extend(handlers)


# -----------------------------------------------------------------------------
# Main entry point
# -----------------------------------------------------------------------------

main = launch_new_instance = JupyterKernelsExtensionApp.launch_instance



# Copyright (c) 2023-2024 Datalayer, Inc.
#
# Datalayer License

"""Jump handler.
"""

import os
import getpass
import json
import random
from pathlib import Path

import tornado
from tornado import concurrent

from jupyter_server.base.handlers import APIHandler
from jupyter_server.extension.handler import ExtensionHandlerMixin

from ...__version__ import __version__


HOME = Path.home()

JUMP_SERVER = "oss.datalayer.run"

LOCAL_USERNAME = getpass.getuser()

# REMOTE_FOLDER = "/home/jovyan"
REMOTE_FOLDER = "/home/jovyan/content/local"

EXECUTOR = concurrent.futures.ThreadPoolExecutor(8)


def mount(pod_name, local_dir, logger):

    port = random.randrange(10000, 40000)

    create_remote_folder_cmd = f"ssh -i ~/.ssh/datalayer-jump -oStrictHostKeyChecking=no -p 2522 jovyan@{pod_name}.jupyter-kernel.datalayer.svc.cluster.local -oProxyCommand='ssh -W %h:%p -p 2223 -i ~/.ssh/datalayer-jump -oStrictHostKeyChecking=no datalayer@{JUMP_SERVER}' 'mkdir {REMOTE_FOLDER}; fusermount -u {REMOTE_FOLDER}; ls {REMOTE_FOLDER}'"
    logger.info(f"Creating remote folder: {create_remote_folder_cmd}")
    os.system(create_remote_folder_cmd)

    # Use this option if you want more debug: -o StrictHostKeyChecking=debug no,debug,sshfs_debug,loglevel=debug,default_permissions,idmap=user
    mount_local_to_remote_folder = f"ssh -R {port}:localhost:22 -i ~/.ssh/datalayer-jump -oStrictHostKeyChecking=no -p 2522 jovyan@{pod_name}.jupyter-kernel.datalayer.svc.cluster.local -oProxyCommand='ssh -W %h:%p -p 2223 -i ~/.ssh/datalayer-jump -oStrictHostKeyChecking=no datalayer@{JUMP_SERVER}' 'sshfs -p {port} -o StrictHostKeyChecking=no,default_permissions,idmap=user {LOCAL_USERNAME}@localhost:{local_dir} {REMOTE_FOLDER}'"
    logger.info(f"Mounting local to remote folder: {mount_local_to_remote_folder}")
    os.system(mount_local_to_remote_folder)


def unmount(pod_name, logger):

#    unmount_cmd = f"ssh -i ~/.ssh/datalayer-jump -oStrictHostKeyChecking=no -p 2522 jovyan@{pod_name}.jupyter-kernel.datalayer.svc.cluster.local -oProxyCommand='ssh -W %h:%p -p 2223 -i ~/.ssh/datalayer-jump -oStrictHostKeyChecking=no datalayer@{JUMP_SERVER}' 'fusermount -u {REMOTE_FOLDER}'"
    unmount_cmd = "pkill ssh"
    logger.info(f"Unmount: {unmount_cmd}")
    os.system(unmount_cmd)


# pylint: disable=W0223
class JumpHandler(ExtensionHandlerMixin, APIHandler):
    """The handler for jump."""

    @tornado.web.authenticated
    @tornado.gen.coroutine
    def get(self, pod_name):
        """Mount a the local folder."""
        root_dir = self.settings.get(
            "server_root_dir", self.settings.get(
              "root_dir", self.settings.get(
                  "notebook_dir", f"{HOME}/Desktop")))
        root_dir_absolute = Path(root_dir).expanduser().resolve()
        self.log.info(f"Requesting connection via Datalayer Jump to pod {pod_name} for local path {root_dir_absolute}")
        EXECUTOR.submit(mount, pod_name, root_dir_absolute, self.log)
        res = json.dumps({
            "extension": "jupyter_kernels",
            "version": __version__,
        })
        self.finish(res)


    @tornado.web.authenticated
    @tornado.gen.coroutine
    def delete(self, pod_name):
        """Unmount a the local folder."""
        self.log.info(f"Requesting unmount via Datalayer Jump to pod {pod_name}")
        EXECUTOR.submit(unmount, pod_name, self.log)
        res = json.dumps({
            "extension": "jupyter_kernels",
            "version": __version__,
        })
        self.finish(res)



# Copyright (c) 2023-2024 Datalayer, Inc.
#
# Datalayer License

"""Config handler."""

import json

import tornado

from jupyter_server.base.handlers import APIHandler
from jupyter_server.extension.handler import ExtensionHandlerMixin

from jupyter_kernels.__version__ import __version__


# pylint: disable=W0223
class ConfigHandler(ExtensionHandlerMixin, APIHandler):
    """The handler for configurations."""

    @tornado.web.authenticated
    def get(self):
        """Returns the configurations of the server extensions."""
        res = json.dumps({
            "extension": "jupyter_kernels",
            "version": __version__,
            "configuration": dict(
                [
                    (
                        ''.join(
                            w.title() if idx > 0 else w
                            for idx, w in enumerate(k.split('_'))
                        ),
                        v
                    ) for k, v in self.settings['jupyter_kernels_config'].items()
                ]
            )
        })
        self.finish(res)
